# Simple example of usage of caqui with Android App
# It opens the app and get the source code
# Appium configured using docker for simplicity
# https://hub.docker.com/r/appium/appium/
# sample app in ./tests/apk folder
from caqui import synchronous


def main():
    driver_url = "http://127.0.0.1:4723"
    capabilities = {
        "capabilities": {
            "firstMatch": [{}],
            "alwaysMatch": {
                "appium:automationName": "UIAutomator2",
                "platformName": "Android",
                "appium:udid": "YOUR-APP-UUID",  # replace with your device uuid
                "appium:app": "/home/androidusr/sample.apk",  # refereces the folder in docker container
            },
        }
    }

    session = synchronous.get_session(driver_url, capabilities)
    print("session: ", session)

    source = synchronous.get_page_source(driver_url, session)
    print("source: ", source)

    synchronous.close_session(driver_url, session)


if __name__ == "__main__":
    main()

# output
# python sample-appium.py
# session:  8264988f-2c63-4935-a46f-3aa8ccc09004
# source:  <?xml version='1.0' encoding='UTF-8' standalone='yes' ?>
# <hierarchy index="0" class="hierarchy" rotation="0" width="720" height="1472">
#   <android.widget.FrameLayout index="0" package="com.example.sample" class="android.widget.FrameLayout" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,0][720,1472]" displayed="true">
#     <android.widget.LinearLayout index="0" package="com.example.sample" class="android.widget.LinearLayout" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,0][720,1472]" displayed="true">
#       <android.widget.FrameLayout index="0" package="com.example.sample" class="android.widget.FrameLayout" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,0][720,1472]" displayed="true">
#         <android.widget.LinearLayout index="0" package="com.example.sample" class="android.widget.LinearLayout" text="" resource-id="com.example.sample:id/action_bar_root" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,0][720,1472]" displayed="true">
#           <android.widget.FrameLayout index="0" package="com.example.sample" class="android.widget.FrameLayout" text="" resource-id="android:id/content" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,0][720,1472]" displayed="true">
#             <android.view.ViewGroup index="0" package="com.example.sample" class="android.view.ViewGroup" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,0][720,1472]" displayed="true">
#               <android.widget.LinearLayout index="0" package="com.example.sample" class="android.widget.LinearLayout" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,0][720,159]" displayed="true">
#                 <android.view.ViewGroup index="0" package="com.example.sample" class="android.view.ViewGroup" text="" resource-id="com.example.sample:id/toolbar" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,47][720,159]" displayed="true">
#                   <android.widget.TextView index="0" package="com.example.sample" class="android.widget.TextView" text="First Fragment" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[28,76][282,129]" displayed="true" />
#                   <androidx.appcompat.widget.LinearLayoutCompat index="1" package="com.example.sample" class="androidx.appcompat.widget.LinearLayoutCompat" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[649,54][720,152]" displayed="true">
#                     <android.widget.ImageView index="0" package="com.example.sample" class="android.widget.ImageView" text="" content-desc="More options" checkable="false" checked="false" clickable="true" enabled="true" focusable="true" focused="false" long-clickable="true" password="false" scrollable="false" selected="false" bounds="[649,61][720,145]" displayed="true" />
#                   </androidx.appcompat.widget.LinearLayoutCompat>
#                 </android.view.ViewGroup>
#               </android.widget.LinearLayout>
#               <android.view.ViewGroup index="1" package="com.example.sample" class="android.view.ViewGroup" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,159][720,1472]" displayed="true">
#                 <android.widget.FrameLayout index="0" package="com.example.sample" class="android.widget.FrameLayout" text="" resource-id="com.example.sample:id/nav_host_fragment_content_main" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,159][720,1472]" displayed="true">
#                   <android.widget.ScrollView index="0" package="com.example.sample" class="android.widget.ScrollView" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="true" focused="false" long-clickable="false" password="false" scrollable="true" selected="false" bounds="[0,159][720,1472]" displayed="true">
#                     <android.view.ViewGroup index="0" package="com.example.sample" class="android.view.ViewGroup" text="" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[0,159][720,1472]" displayed="true">
#                       <android.widget.Button index="0" package="com.example.sample" class="android.widget.Button" text="Next" resource-id="com.example.sample:id/button_first" checkable="false" checked="false" clickable="true" enabled="true" focusable="true" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[283,187][437,271]" displayed="true" />
#                       <android.widget.TextView index="1" package="com.example.sample" class="android.widget.TextView" text="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam in scelerisque sem. Mauris volutpat, dolor id interdum ullamcorper, risus dolor egestas lectus, sit amet mattis purus dui nec risus. Maecenas non sodales nisi, vel dictum dolor. Class aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos himenaeos. Suspendisse blandit eleifend diam, vel rutrum tellus vulputate quis. Aliquam eget libero aliquet, imperdiet nisl a, ornare ex. Sed rhoncus est ut libero porta lobortis. Fusce in dictum tellus.&#10;&#10; Suspendisse interdum ornare ante. Aliquam nec cursus lorem. Morbi id magna felis. Vivamus egestas, est a condimentum egestas, turpis nisl iaculis ipsum, in dictum tellus dolor sed neque. Morbi tellus erat, dapibus ut sem a, iaculis tincidunt dui. Interdum et malesuada fames ac ante ipsum primis in faucibus. Curabitur et eros porttitor, ultricies urna vitae, molestie nibh. Phasellus at commodo eros, non aliquet metus. Sed maximus nisl nec dolor bibendum, vel congue leo egestas.&#10;&#10; Sed interdum tortor nibh, in sagittis risus mollis quis. Curabitur mi odio, condimentum sit amet auctor at, mollis non turpis. Nullam pretium libero vestibulum, finibus orci vel, molestie quam. Fusce blandit tincidunt nulla, quis sollicitudin libero facilisis et. Integer interdum nunc ligula, et fermentum metus hendrerit id. Vestibulum lectus felis, dictum at lacinia sit amet, tristique id quam. Cras eu consequat dui. Suspendisse sodales nunc ligula, in lobortis sem porta sed. Integer id ultrices magna, in luctus elit. Sed a pellentesque est.&#10;&#10; Aenean nunc velit, lacinia sed dolor sed, ultrices viverra nulla. Etiam a venenatis nibh. Morbi laoreet, tortor sed facilisis varius, nibh orci rhoncus nulla, id elementum leo dui non lorem. Nam mollis ipsum quis auctor varius. Quisque elementum eu libero sed commodo. In eros nisl, imperdiet vel imperdiet et, scelerisque a mauris. Pellentesque varius ex nunc, quis imperdiet eros placerat ac. Duis finibus orci et est auctor tincidunt. Sed non viverra ipsum. Nunc quis augue egestas, cursus lorem at, molestie sem. Morbi a consectetur ipsum, a placerat diam. Etiam vulputate dignissim convallis. Integer faucibus mauris sit amet finibus convallis.&#10;&#10; Phasellus in aliquet mi. Pellentesque habitant morbi tristique senectus et netus et malesuada fames ac turpis egestas. In volutpat arcu ut felis sagittis, in finibus massa gravida. Pellentesque id tellus orci. Integer dictum, lorem sed efficitur ullamcorper, libero justo consectetur ipsum, in mollis nisl ex sed nisl. Donec maximus ullamcorper sodales. Praesent bibendum rhoncus tellus nec feugiat. In a ornare nulla. Donec rhoncus libero vel nunc consequat, quis tincidunt nisl eleifend. Cras bibendum enim a justo luctus vestibulum. Fusce dictum libero quis erat maximus, vitae volutpat diam dignissim." resource-id="com.example.sample:id/textview_first" checkable="false" checked="false" clickable="false" enabled="true" focusable="false" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[28,299][692,1472]" displayed="true" />
#                     </android.view.ViewGroup>
#                   </android.widget.ScrollView>
#                 </android.widget.FrameLayout>
#               </android.view.ViewGroup>
#               <android.widget.ImageButton index="2" package="com.example.sample" class="android.widget.ImageButton" text="" resource-id="com.example.sample:id/fab" checkable="false" checked="false" clickable="true" enabled="true" focusable="true" focused="false" long-clickable="false" password="false" scrollable="false" selected="false" bounds="[594,1390][692,1472]" displayed="true" />
#             </android.view.ViewGroup>
#           </android.widget.FrameLayout>
#         </android.widget.LinearLayout>
#       </android.widget.FrameLayout>
#     </android.widget.LinearLayout>
#   </android.widget.FrameLayout>
# </hierarchy>



#### File used to figure out requests format and parameters ####

from selenium import webdriver
from pytest import fixture, mark
from tests.constants import PAGE_URL
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support import expected_conditions, wait
from selenium.webdriver.common.alert import Alert
import os


@fixture
def setup():
    desired_capabilities = {
        # 'deviceName': 'Device',
        "deviceName": "Emulator",
        "deviceIpAddress": "127.0.0.1",
        "locale": "en-US",
        "debugCodedUI": False,
        "app": "chromedriver",
    }

    driver = webdriver.Remote(
        command_executor="http://localhost:9999",
        desired_capabilities=desired_capabilities,
    )
    driver.get(PAGE_URL)
    driver
    yield driver
    driver.quit()


@fixture
def setup_binary():
    homedir = os.path.expanduser("~")
    service = Service(f"/home/douglas/web_drivers/chromedriver.113")
    options = Options()
    # options.add_argument('--headless')
    options.add_argument("window-size=1920,1080")
    browser = webdriver.Chrome(service=service, options=options)
    browser.get(PAGE_URL)

    yield browser
    # browser.quit()


@mark.skip("used just to discover request data")
def test_switch_to_frame_sniffer(setup):
    driver = setup
    # Click the link to activate the alert
    import time

    time.sleep(3)
    element = driver.find_element(By.ID, "my-iframe")
    driver.switch_to.frame(element)

    element_alert = driver.find_element(By.ID, "alert-button-iframe")
    element_alert.click()
    time.sleep(3)


@mark.skip("used just to discover request data")
def test_send_text_to_alert(setup):
    driver = setup
    # Click the link to activate the alert
    driver.find_element(By.CSS_SELECTOR, "#alert-button-prompt").click()
    import time

    time.sleep(3)

    # Wait for the alert to be displayed
    # wait.until(expected_conditions.alert_is_present())

    # Store the alert in a variable for reuse
    alert = Alert(driver)
    print("alert ", alert.text)

    # Type your message
    alert.send_keys("Selenium")
    time.sleep(3)

    # Press the OK button
    alert.accept()
    time.sleep(3)


@mark.skip("used just to discover request data")
def test_add_cookie(setup):
    driver = setup
    driver.add_cookie({"name": "firstname", "value": "James"})


@mark.skip("used just to discover request data")
def test_submit(setup):
    driver = setup
    search_button = driver.find_element("name", "my-form")
    search_button.submit()


@mark.skip("used just to discover request data")
def test_sniffer_is_displayed(setup_binary):
    driver = setup_binary
    search_button = driver.find_element("xpath", "//button")
    search_button.is_displayed()


@mark.skip("used just to discover request data")
def test_sniffer_actions_scroll_to_element(setup_binary):
    driver = setup_binary
    search_button = driver.find_element("xpath", "//button")
    ActionChains(driver).scroll_to_element(search_button).perform()


@mark.skip("used just to discover request data")
def test_sniffer_actions_click(setup):
    driver = setup
    search_button = driver.find_element("xpath", "//button")
    ActionChains(driver).click(search_button).perform()


@mark.skip("used just to discover request data")
def test_sniffer_find_children_elements(setup):
    driver = setup
    element = driver.find_element(By.XPATH, '//div[@class="parent"]')
    elements = element.find_elements(By.XPATH, "//div")
    for e in elements:
        print("element_text:", e.text)


@mark.skip("used just to discover request data")
def test_exec_script(setup):
    driver = setup
    assert driver.execute_script("return document.body.scrollHeight") == "any"


@mark.skip("used just to discover request data")
def test_clear(setup):
    driver = setup
    element = driver.find_element("xpath", "//input")
    assert element.clear() == "any"


@mark.skip("used just to discover request data")
def test_back(setup):
    assert setup.back() == "any"


@mark.skip("used just to discover request data")
def test_get_title(setup):
    assert setup.title == "any"


@mark.skip("used just to discover request data")
def test_get_attribute(setup):
    driver = setup
    element = driver.find_element("xpath", "//a")
    assert element.get_property("href") == "any"


@mark.skip("used just to discover request data")
def test_find_elements(setup):
    driver = setup
    actual = driver.find_elements("xpath", "//input")
    assert actual == "any"


@mark.skip("used just to discover request data")
def test_get_attribute_from_input(setup):
    driver = setup
    search_box = driver.find_element("xpath", "//input")
    search_box.send_keys("cat")

    data = driver.find_element("xpath", "//input").get_property("value")
    assert data == "cat"


@mark.skip("used just to discover request data")
def test_sniff(setup):
    driver = setup
    search_box = driver.find_element("xpath", "//input")
    search_button = driver.find_element("xpath", "//button")
    end = driver.find_element("xpath", "//p[@id='end']")
    search_box.send_keys("cat")
    search_button.click()
    assert end.text == "end"



import aiohttp
import json
from caqui.constants import HEADERS
from caqui.exceptions import WebDriverError
from caqui import helper


async def __handle_response(resp):
    result = None
    if resp.status in range(200, 399):
        result = await resp.json()
    else:
        raise WebDriverError(f"Status code: {resp.status}, Body: {resp.text}")

    if int(result.get("status", 0)) > 0:
        raise WebDriverError(f"Status code: {resp.status}, Body: {resp.text}")

    return result


async def __delete(url):
    try:
        async with aiohttp.ClientSession() as session:
            async with session.delete(url, headers=HEADERS) as resp:
                return await __handle_response(resp)
    except Exception as error:
        raise WebDriverError("'DELETE' request failed.") from error


async def __post(url, payload):
    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url, data=json.dumps(payload), headers=HEADERS
            ) as resp:
                return await __handle_response(resp)
    except Exception as error:
        raise WebDriverError("'POST' request failed.") from error


async def __get(url):
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=HEADERS) as resp:
                return await __handle_response(resp)
    except Exception as error:
        raise WebDriverError("'GET' request failed.") from error


async def __handle_alert(driver_url, session, command):
    url = f"{driver_url}/session/{session}/alert/{command}"
    payload = {
        "value": command,
    }
    await __post(url, payload)
    return True


async def switch_to_parent_frame(driver_url, session, element_frame):
    """Switch to parent frame of 'element_frame'"""
    try:
        url = f"{driver_url}/session/{session}/frame/parent"
        payload = {"id": {"ELEMENT": element_frame}}
        await __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to switch to parent frame.") from error


async def switch_to_frame(driver_url, session, element_frame):
    """Switch to frame 'element_frame'"""
    try:
        url = f"{driver_url}/session/{session}/frame"
        payload = {"id": {"ELEMENT": element_frame}}
        await __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to switch to frame.") from error


async def delete_all_cookies(driver_url, session):
    """Delete all cookies"""
    try:
        url = f"{driver_url}/session/{session}/cookie"
        await __delete(url)
        return True
    except Exception as error:
        raise WebDriverError("Failed to delete cookies.") from error


async def send_alert_text(driver_url, session, text):
    """Fill the alert text area and send the text"""
    try:
        url = f"{driver_url}/session/{session}/alert/text"
        payload = {
            "text": text,
        }
        await __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to sent text to alert.") from error


async def accept_alert(driver_url, session):
    """Accept alert"""
    try:
        return await __handle_alert(driver_url, session, "accept")
    except Exception as error:
        raise WebDriverError(f"Failed to accept alert.") from error


async def dismiss_alert(driver_url, session):
    """Dismiss alert"""
    try:
        return await __handle_alert(driver_url, session, "dismiss")
    except Exception as error:
        raise WebDriverError(f"Failed to dismiss alert.") from error


async def take_screenshot_element(
    driver_url, session, element, path="/tmp", file_name="caqui"
):
    """Take screenshot of element"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/screenshot"
        response = await __get(url)
        picture = response.get("value")
        helper.save_picture(session, path, file_name, picture)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to take screenshot.") from error


async def take_screenshot(driver_url, session, path="/tmp", file_name="caqui"):
    """Take screenshot"""
    try:
        url = f"{driver_url}/session/{session}/screenshot"
        response = await __get(url)
        picture = response.get("value")
        helper.save_picture(session, path, file_name, picture)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to take screenshot.") from error


async def get_named_cookie(driver_url, session, name):
    """Get cookie by name"""
    try:
        url = f"{driver_url}/session/{session}/cookie/{name}"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get cookie '{name}'.") from error


async def get_computed_label(driver_url, session, element):
    """Get the element tag computed label. Get the accessibility name"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/computedlabel"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get element computed label.") from error


async def get_computed_role(driver_url, session, element):
    """Get the element tag computed role (the element role)"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/computedrole"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get element computed role.") from error


async def get_tag_name(driver_url, session, element):
    """Get the element tag name"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/name"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get element name.") from error


async def get_shadow_root(driver_url, session, element):
    """Get the shadow root element"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/shadow"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get element shadow.") from error


async def get_rect(driver_url, session, element):
    """Get the element rectangle"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/rect"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get element rect.") from error


async def actions(driver_url, session, payload):
    url = f"{driver_url}/session/{session}/actions"
    await __post(url, payload)
    return True


async def actions_scroll_to_element(driver_url, session, element):
    """Scroll to an element simulating a mouse movement"""
    try:
        payload = {
            "actions": [
                {
                    "type": "wheel",
                    "id": "wheel",
                    "actions": [
                        {
                            "type": "scroll",
                            "x": 0,
                            "y": 0,
                            "deltaX": 0,
                            "deltaY": 0,
                            "duration": 0,
                            "origin": {"ELEMENT": element},
                        }
                    ],
                }
            ]
        }
        return await actions(driver_url, session, payload)
    except Exception as error:
        raise WebDriverError(f"Failed to scroll to element.") from error


async def submit(driver_url, session, element):
    """Submit a form. It is similar to 'submit' funtion in Seleniu
    It is not part of W3C WebDriver. Just added for convenience
    """
    try:
        submit_element = await find_child_element(
            driver_url,
            session,
            element,
            locator_type="xpath",
            locator_value="*[@type='submit']",
        )
        return await click(driver_url, session, submit_element)
    except Exception as error:
        raise WebDriverError(f"Failed to submit form.") from error


async def actions_click(driver_url, session, element):
    """Click an element simulating a mouse movement"""
    try:
        payload = {
            "actions": [
                {
                    "type": "pointer",
                    "parameters": {"pointerType": "mouse"},
                    "id": "mouse",
                    "actions": [
                        {
                            "type": "pointerMove",
                            "duration": 250,
                            "x": 0,
                            "y": 0,
                            "origin": {"ELEMENT": element},
                        },
                        {"type": "pointerDown", "duration": 0, "button": 0},
                        {"type": "pointerUp", "duration": 0, "button": 0},
                    ],
                },
                {
                    "type": "key",
                    "id": "key",
                    "actions": [
                        {"type": "pause", "duration": 0},
                        {"type": "pause", "duration": 0},
                        {"type": "pause", "duration": 0},
                    ],
                },
            ]
        }
        return await actions(driver_url, session, payload)
    except Exception as error:
        raise WebDriverError(f"Failed to click the element.") from error


async def set_timeouts(driver_url, session, timeouts):
    """Set timeouts"""
    try:
        url = f"{driver_url}/session/{session}/timeouts"
        payload = {
            "implicit": timeouts,
        }
        await __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to set timeouts.") from error


async def find_children_elements(
    driver_url, session, parent_element, locator_type, locator_value
):
    """Find the children elements by 'locator_type'"""
    try:
        url = f"{driver_url}/session/{session}/element/{parent_element}/elements"
        payload = {"using": locator_type, "value": locator_value, "id": parent_element}
        response = await __post(url, payload)
        return helper.get_elements(response)
    except Exception as error:
        raise WebDriverError(
            f"Failed to find the children elements from '{parent_element}'."
        ) from error


async def find_child_element(
    driver_url, session, parent_element, locator_type, locator_value
):
    """Find the child element by 'locator_type'"""
    try:
        url = f"{driver_url}/session/{session}/element/{parent_element}/element"
        payload = {"using": locator_type, "value": locator_value, "id": parent_element}
        response = await __post(url, payload)
        return helper.get_element(response)
    except Exception as error:
        raise WebDriverError(
            f"Failed to find the child element from '{parent_element}'."
        ) from error


async def get_page_source(driver_url, session):
    """Get the page source (all content)"""
    try:
        url = f"{driver_url}/session/{session}/source"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get the page source.") from error


async def execute_script(driver_url, session, script, args=[]):
    """Executes a script, like 'alert('something')' to open an alert window"""
    try:
        url = f"{driver_url}/session/{session}/execute/sync"
        payload = {"script": script, "args": args}
        response = await __post(url, payload)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to open session.") from error


async def get_alert_text(driver_url, session):
    """Get the text from an alert"""
    try:
        url = f"{driver_url}/session/{session}/alert/text"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get the alert text.") from error


async def get_active_element(driver_url, session):
    """Get the active element"""
    try:
        url = f"{driver_url}/session/{session}/element/active"
        response = await __get(url)
        return helper.get_element(response)
    except Exception as error:
        raise WebDriverError("Failed to check if element is selected.") from error


async def clear_element(driver_url, session, element):
    """Clear the element text"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/clear"
        payload = {"id": element}
        await __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError("Failed to clear the element text.") from error


async def is_element_enabled(driver_url, session, element):
    """Check if element is enabled"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/enabled"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to check if element is enabled.") from error


async def get_css_value(driver_url, session, element, property_name):
    """Check if element is selected"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/css/{property_name}"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to check if element is selected.") from error


async def is_element_selected(driver_url, session, element):
    """Check if element is selected"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/selected"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to check if element is selected.") from error


async def get_window_rectangle(driver_url, session):
    """Get window rectangle"""
    try:
        url = f"{driver_url}/session/{session}/window/rect"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get window rectangle.") from error


async def get_window_handles(driver_url, session):
    """Get window handles"""
    try:
        url = f"{driver_url}/session/{session}/window/handles"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get window handles.") from error


async def close_window(driver_url, session):
    """Close active window"""
    try:
        url = f"{driver_url}/session/{session}/window"
        response = await __delete(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to close active window.") from error


async def get_window(driver_url, session):
    """Get window"""
    try:
        url = f"{driver_url}/session/{session}/window"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get window.") from error


async def go_back(driver_url, session):
    """
    This command causes the browser to traverse one step backward in the joint session history of the
    current browse. This is equivalent to pressing the back button in the browser.
    """
    try:
        url = f"{driver_url}/session/{session}/back"
        await __post(url, {})
        return True
    except Exception as error:
        raise WebDriverError("Failed to go back to page.") from error


async def get_url(driver_url, session):
    """Return the URL from web page:"""
    try:
        url = f"{driver_url}/session/{session}/url"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get page url.") from error


async def get_timeouts(driver_url, session):
    """
    Return the configured timeouts:
        {"implicit": 0, "pageLoad": 300000, "script": 30000}
    """
    try:
        url = f"{driver_url}/session/{session}/timeouts"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get timeouts.") from error


async def get_status(driver_url):
    """
    Return the status and details of the WebDriver:
        "build": {
                "version": "113.0.5672.63 (0e1a4471d5ae5bf128b1bd8f4d627c8cbd55f70c-refs/branch-heads/5672@{#912})"
            },
            "message": "ChromeDriver ready for new sessions.",
            "os": {"arch": "x86_64", "name": "Linux", "version": "5.4.0-150-generic"},
            "ready": True,
        }
    """
    try:
        url = f"{driver_url}/status"
        response = await __get(url)
        return response
    except Exception as error:
        raise WebDriverError("Failed to get status.") from error


async def get_title(driver_url, session):
    """Get the page title"""
    try:
        url = f"{driver_url}/session/{session}/title"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get page title.") from error


async def find_elements(driver_url, session, locator_type, locator_value):
    """Search the DOM elements by 'locator', for example, 'xpath'"""
    try:
        payload = {"using": locator_type, "value": locator_value}
        url = f"{driver_url}/session/{session}/elements"
        response = await __post(url, payload)
        return [x.get("ELEMENT") for x in response.get("value")]
    except Exception as error:
        raise WebDriverError(
            f"Failed to find element by '{locator_type}'-'{locator_value}'."
        ) from error


async def get_property(driver_url, session, element, property):
    """Get the given HTML property of an element, for example, 'href'"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/property/{property}"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get value from element.") from error


async def get_attribute(driver_url, session, element, attribute):
    """Get the given HTML attribute of an element, for example, 'aria-valuenow'"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/attribute/{attribute}"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get value from element.") from error


async def get_text(driver_url, session, element):
    """Get the text of an element"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/text"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get text from element.") from error


async def get_cookies(driver_url, session):
    """Get the page cookies"""
    try:
        url = f"{driver_url}/session/{session}/cookie"
        response = await __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get page cookies.") from error


async def close_session(driver_url, session):
    """Close an opened session and close the browser"""
    try:
        url = f"{driver_url}/session/{session}"
        await __delete(url)
        return True
    except Exception as error:
        raise WebDriverError("Failed to close session.") from error


async def get(driver_url, session, page_url):
    """Does the same of 'go_to_page'. Added to be compatible with selenium method name'"""
    return go_to_page(driver_url, session, page_url)


async def go_to_page(driver_url, session, page_url):
    """Navigate to 'page_url'"""
    try:
        url = f"{driver_url}/session/{session}/url"
        payload = {"url": page_url}
        await __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to navigate to page '{page_url}'.") from error


async def send_keys(driver_url, session, element, text):
    """Fill an editable element, for example a textarea, with a given text"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/value"
        payload = {"text": text, "value": [*text], "id": element}
        await __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to send key '{text}'.") from error


async def click(driver_url, session, element):
    """Click on an element"""
    try:
        payload = {"id": element}
        url = f"{driver_url}/session/{session}/element/{element}/click"
        await __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError("Failed to click on element.") from error


async def find_element(driver_url, session, locator_type, locator_value):
    """Find an element by a 'locator', for example 'xpath'"""

    try:
        payload = {"using": locator_type, "value": locator_value}
        url = f"{driver_url}/session/{session}/element"
        response = await __post(url, payload)

        # Firefox does not support id locator, so it prints the error message to the user
        # It helps on debug
        if response.get("value").get("error"):
            raise WebDriverError(f"Failed to find element. {response}")
        return helper.get_element(response)
    except Exception as error:
        raise WebDriverError(
            f"Failed to find element by '{locator_type}'-'{locator_value}'."
        ) from error


async def get_session(driver_url, capabilities):
    """Opens a browser and a session. This session is used for all functions to perform events in the page"""
    try:
        payload = capabilities
        url = f"{driver_url}/session"
        response = await __post(url, payload)
        return response.get("sessionId")
    except Exception as error:
        raise WebDriverError("Failed to open session.") from error



from unittest.mock import patch
from caqui import synchronous
from tests import fake_responses


@patch("requests.request", return_value=fake_responses.GET_RECT)
def test_get_rect(*args):
    expected = {"height": 23, "width": 183, "x": 10, "y": 9652.12}
    assert synchronous.get_rect("", "", "") == expected


@patch("requests.request", return_value=fake_responses.ACTIONS)
def test_actions_scroll_to_element(*args):
    assert synchronous.actions_scroll_to_element("", "", "") == True


@patch("requests.request", return_value=fake_responses.CLICK)
def test_submit(*args):
    assert synchronous.submit("", "", "") == True


@patch("requests.request", return_value=fake_responses.ACTIONS)
def test_actions_click(*args):
    assert synchronous.actions_click("", "", "") == True


@patch("requests.request", return_value=fake_responses.GET_TIMEOUTS)
def test_set_timeouts(*args):
    assert synchronous.set_timeouts("", "", "") == True


@patch("requests.request", return_value=fake_responses.FIND_ELEMENTS)
def test_find_children_elements(*args):
    element = "C230605181E69CB2C4C36B8E83FE1245_element_2"

    elements = synchronous.find_children_elements("", "", "", "", "")

    assert element in elements
    assert len(elements) == 3


@patch("requests.request", return_value=fake_responses.FIND_ELEMENT)
def test_find_child_element(*args):
    expected = "0.8851292311864847-1"

    assert synchronous.find_child_element("", "", "", "", "") == expected


@patch("requests.request", return_value=fake_responses.EXECUTE_SCRIPT)
def test_execute_script(*args):
    expected = "any"

    assert synchronous.execute_script("", "", "", "") == expected


@patch("requests.request", return_value=fake_responses.GET_PAGE_SOURCE)
def test_get_page_source(*args):
    expected = "Sample page"
    assert expected in synchronous.get_page_source("", "")


@patch("requests.request", return_value=fake_responses.GET_ALERT_TEXT)
def test_get_alert_text(*args):
    expected = "any warn"
    assert synchronous.get_alert_text("", "") == expected


@patch("requests.request", return_value=fake_responses.GET_ACTIVE_ELEMENT)
def test_get_active_element(*args):
    expected = "0.8851292311864847-1"
    assert synchronous.get_active_element("", "") == expected


@patch("requests.request", return_value=fake_responses.CLEAR_ELEMENT)
def test_clear_element(*args):
    assert synchronous.clear_element("", "", "") is True


@patch("requests.request", return_value=fake_responses.IS_ELEMENT_ENABLED)
def test_is_element_enabled(*args):
    assert synchronous.is_element_enabled("", "", "") is True


@patch("requests.request", return_value=fake_responses.GET_CSS_COLOR_VALUE)
def test_get_css_value(*args):
    expected = "rgba(0, 0, 0, 1)"
    assert synchronous.get_css_value("", "", "", "") == expected


@patch("requests.request", return_value=fake_responses.IS_ELEMENT_SELECTED)
def test_is_element_selected(*args):
    assert synchronous.is_element_selected("", "", "") is False


@patch("requests.request", return_value=fake_responses.GET_WINDOW_RECTANGLE)
def test_get_window_rectangle(*args):
    expected = "height"

    assert expected in synchronous.get_window_rectangle("", "")


@patch("requests.request", return_value=fake_responses.GET_WINDOW_HANDLES)
def test_get_window_handles(*args):
    expected = "2E55CCE389196328988ED244DAA52A5D"

    assert expected in synchronous.get_window_handles("", "")


@patch("requests.request", return_value=fake_responses.CLOSE_WINDOW)
def test_close_window(*args):
    expected = []

    assert synchronous.close_window("", "") == expected


@patch("requests.request", return_value=fake_responses.GET_WINDOW)
def test_get_window(*args):
    expected = "845623CAE8115F2B60C9AE8596F13D94"

    assert expected in synchronous.get_window("", "")


@patch("requests.request", return_value=fake_responses.GET_URL)
def test_get_url(*args):
    expected = "playground.html"

    assert expected in synchronous.get_url("", "")


@patch("requests.request", return_value=fake_responses.GET_TIMEOUTS)
def test_get_timeouts(*args):
    expected = "implicit"

    assert expected in synchronous.get_timeouts("", "")


@patch("requests.request", return_value=fake_responses.GET_STATUS)
def test_get_status(*args):
    assert synchronous.get_status("").get("value").get("ready") is True


@patch("requests.request", return_value=fake_responses.GET_TITLE)
def test_get_title(*args):
    expected = "Sample page"

    assert synchronous.get_title("", "") == expected


@patch("requests.request", return_value=fake_responses.GET_COOKIES)
def test_get_cookies(*args):
    expected = []

    assert synchronous.get_cookies("", "") == expected


@patch("requests.request", return_value=fake_responses.FIND_ELEMENTS)
def test_find_elements(*args):
    element = "C230605181E69CB2C4C36B8E83FE1245_element_2"

    elements = synchronous.find_elements("", "", "", "")

    assert element in elements
    assert len(elements) == 3


@patch("requests.request", return_value=fake_responses.GET_PROPERTY_VALUE)
def test_get_property(*args):
    expected = "any_value"

    assert synchronous.get_property("", "", "", "") == expected


@patch("requests.request", return_value=fake_responses.GET_ATTRIBUTE_VALUE)
def test_get_attribute(*args):
    expected = "any_value"

    assert synchronous.get_attribute("", "", "", "") == expected


@patch("requests.request", return_value=fake_responses.GO_TO_PAGE)
def test_go_to_page(*args):
    assert synchronous.go_to_page("", "", "") is True


@patch("requests.request", return_value=fake_responses.CLOSE_SESSION)
def test_close_session(*args):
    assert synchronous.close_session("", "") is True


@patch("requests.request", return_value=fake_responses.GET_TEXT)
def test_get_text(*args):
    expected = "any"

    assert synchronous.get_text("", "", "") == expected


@patch("requests.request", return_value=fake_responses.SEND_KEYS)
def test_send_keys(*args):
    assert synchronous.send_keys("", "", "", "") is True


@patch("requests.request", return_value=fake_responses.CLICK)
def test_click(*args):
    assert synchronous.click("", "", "") is True


@patch("requests.request", return_value=fake_responses.GET_SESSION)
def test_get_session(*args):
    expected = "4358a5b53794586af59678fc1653dc40"

    assert synchronous.get_session("", "") == expected


@patch("requests.request", return_value=fake_responses.FIND_ELEMENT)
def test_find_element(*args):
    expected = "0.8851292311864847-1"

    assert synchronous.find_element("", "", "", "") == expected



from caqui import synchronous, asynchronous
from tests.constants import PAGE_URL
from pytest import fixture, mark


@fixture
def __setup():
    driver_url = "http://127.0.0.1:9999"
    capabilities = {
        "desiredCapabilities": {
            "browserName": "firefox",
            "marionette": True,
            "acceptInsecureCerts": True,
            "goog:chromeOptions": {"extensions": [], "args": ["--headless"]},
        }
    }
    session = synchronous.get_session(driver_url, capabilities)
    synchronous.go_to_page(
        driver_url,
        session,
        PAGE_URL,
    )
    yield driver_url, session
    synchronous.close_session(driver_url, session)


@mark.asyncio
async def test_get_all_links(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    anchors = []

    for i in range(4):
        i += 1
        locator_value = f"//a[@id='a{i}']"
        anchor = synchronous.find_element(
            driver_url, session, locator_type, locator_value
        )
        anchors.append(anchor)
        assert (
            await asynchronous.get_text(driver_url, session, anchors[i - 1])
            == f"any{i}.com"
        )



import requests
import json
from caqui.exceptions import WebDriverError
from caqui import helper

HEADERS = {
    "Accept-Encoding": "identity",
    "Accept": "application/json",
    "Content-Type": "application/json;charset=UTF-8",
    "Connection": "keep-alive",
}


def __handle_response(response):
    result = None
    if response.status_code in range(200, 399):
        result = response.json()
    else:
        raise WebDriverError(f"Status: {response.status_code}, {response.text}")

    if int(result.get("status", 0)) > 0:
        raise WebDriverError(f"Status: {response.status_code}, {response.text}")
    return result


def __get(url):
    try:
        response = requests.request("GET", url, headers=HEADERS, data={})
        return __handle_response(response)
    except Exception as error:
        raise WebDriverError("'GET' request failed.") from error


def __post(url, payload):
    response = requests.request("POST", url, headers=HEADERS, data=json.dumps(payload))
    try:
        return __handle_response(response)
    except Exception as error:
        raise WebDriverError("'POST' request failed.") from error


def __delete(url):
    try:
        response = requests.request("DELETE", url, headers={}, data={})
        return __handle_response(response)
    except Exception as error:
        raise WebDriverError("'DELETE' request failed.") from error


def __handle_alerts(driver_url, session, command):
    url = f"{driver_url}/session/{session}/alert/{command}"
    payload = {"value": command}
    __post(url, payload)
    return True


# def add_cookie(driver_url, session, name, value):
#     """Add cookie by name"""
#     try:
#         url = f"{driver_url}/session/{session}/cookie"
#         payload = {"cookie": {"name": name, "value": value}}
#         __post(url, payload)
#         return True
#     except Exception as error:
#         raise WebDriverError(f"Failed to add cookie '{name}'.") from error


def switch_to_parent_frame(driver_url, session, element_frame):
    """Switch to parent frame of 'element_frame'"""
    try:
        url = f"{driver_url}/session/{session}/frame/parent"
        payload = {"id": {"ELEMENT": element_frame}}
        __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to switch to parent frame.") from error


def switch_to_frame(driver_url, session, element_frame):
    """Switch to frame 'element_frame'"""
    try:
        url = f"{driver_url}/session/{session}/frame"
        payload = {"id": {"ELEMENT": element_frame}}
        __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to switch to frame.") from error


def delete_all_cookies(driver_url, session):
    """Delete all cookies"""
    try:
        url = f"{driver_url}/session/{session}/cookie"
        __delete(url)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to delete cookies.") from error


def send_alert_text(driver_url, session, text):
    """Fill the alert text area and send the text"""
    try:
        url = f"{driver_url}/session/{session}/alert/text"
        payload = {"text": text}
        __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to sent text to alert.") from error


def accept_alert(driver_url, session):
    """Accept an alert"""
    try:
        return __handle_alerts(driver_url, session, "accept")
    except Exception as error:
        raise WebDriverError(f"Failed to accept the alert.") from error


def dismiss_alert(driver_url, session):
    """Dismiss an alert"""
    try:
        return __handle_alerts(driver_url, session, "dismiss")
    except Exception as error:
        raise WebDriverError(f"Failed to dismiss the alert.") from error


def take_screenshot_element(
    driver_url, session, element, path="/tmp", file_name="caqui"
):
    """Take screenshot of element."""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/screenshot"
        response = __get(url).get("value")
        helper.save_picture(session, path, file_name, response)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to take screeshot.") from error


def take_screenshot(driver_url, session, path="/tmp", file_name="caqui"):
    """Take screenshot."""
    try:
        url = f"{driver_url}/session/{session}/screenshot"
        response = __get(url).get("value")
        helper.save_picture(session, path, file_name, response)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to take screeshot.") from error


def get_named_cookie(driver_url, session, name):
    """Get cookie by name."""
    try:
        url = f"{driver_url}/session/{session}/cookie/{name}"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the cookie '{name}'.") from error


def get_computed_label(driver_url, session, element):
    """Get the element computed label. Get the accessibility name."""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/computedlabel"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the element computed label.") from error


def get_computed_role(driver_url, session, element):
    """Get the element computed role (the element role)"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/computedrole"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the element computed role.") from error


def get_tag_name(driver_url, session, element):
    """Get the element tag name"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/name"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the element name.") from error


def get_shadow_root(driver_url, session, element):
    """Get the shadow root element"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/shadow"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the element shadow.") from error


def get_rect(driver_url, session, element):
    """Get the element rectangle"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/rect"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the element rect.") from error


def actions_scroll_to_element(driver_url, session, element):
    """Scroll to an element simulating a mouse movement"""
    try:
        payload = {
            "actions": [
                {
                    "type": "wheel",
                    "id": "wheel",
                    "actions": [
                        {
                            "type": "scroll",
                            "x": 0,
                            "y": 0,
                            "deltaX": 0,
                            "deltaY": 0,
                            "duration": 0,
                            "origin": {"ELEMENT": element},
                        }
                    ],
                }
            ]
        }
        return actions(driver_url, session, payload)
    except Exception as error:
        raise WebDriverError(f"Failed to scroll to element.") from error


def actions(driver_url, session, payload):
    url = f"{driver_url}/session/{session}/actions"
    __post(url, payload)
    return True


def submit(driver_url, session, element):
    """Submit a form. It is similar to 'submit' funtion in Seleniu
    It is not part of W3C WebDriver. Just added for convenience
    """
    try:
        submit_element = find_child_element(
            driver_url,
            session,
            element,
            locator_type="xpath",
            locator_value="//*[@type='submit']",
        )
        return click(driver_url, session, submit_element)
    except Exception as error:
        raise WebDriverError(f"Failed to submit form.") from error


def actions_click(driver_url, session, element):
    """Click an element simulating a mouse movement"""
    try:
        payload = {
            "actions": [
                {
                    "type": "pointer",
                    "parameters": {"pointerType": "mouse"},
                    "id": "mouse",
                    "actions": [
                        {
                            "type": "pointerMove",
                            "duration": 250,
                            "x": 0,
                            "y": 0,
                            "origin": {"ELEMENT": element},
                        },
                        {"type": "pointerDown", "duration": 0, "button": 0},
                        {"type": "pointerUp", "duration": 0, "button": 0},
                    ],
                },
                {
                    "type": "key",
                    "id": "key",
                    "actions": [
                        {"type": "pause", "duration": 0},
                        {"type": "pause", "duration": 0},
                        {"type": "pause", "duration": 0},
                    ],
                },
            ]
        }
        return actions(driver_url, session, payload)
    except Exception as error:
        raise WebDriverError(f"Failed to click the element.") from error


def set_timeouts(driver_url, session, timeouts):
    """Set timeouts"""
    try:
        url = f"{driver_url}/session/{session}/timeouts"
        payload = {
            "implicit": timeouts,
        }
        __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to set timeouts.") from error


def find_children_elements(
    driver_url, session, parent_element, locator_type, locator_value
):
    """Find the children elements by 'locator_type'"""
    try:
        url = f"{driver_url}/session/{session}/element/{parent_element}/elements"
        payload = {"using": locator_type, "value": locator_value, "id": parent_element}
        response = __post(url, payload)
        return helper.get_elements(response)
    except Exception as error:
        raise WebDriverError(
            f"Failed to find the children elements from '{parent_element}'."
        ) from error


def find_child_element(
    driver_url, session, parent_element, locator_type, locator_value
):
    """Find the child element by 'locator_type'"""
    try:
        url = f"{driver_url}/session/{session}/element/{parent_element}/element"
        payload = {"using": locator_type, "value": locator_value, "id": parent_element}
        response = __post(url, payload)
        return helper.get_element(response)
    except Exception as error:
        raise WebDriverError(
            f"Failed to find the child element from '{parent_element}'."
        ) from error


def get_page_source(driver_url, session):
    """Get the page source (all content)"""
    try:
        url = f"{driver_url}/session/{session}/source"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the page source.") from error


def execute_script(driver_url, session, script, args=[]):
    """Executes a script, like 'alert('something')' to open an alert window"""
    try:
        url = f"{driver_url}/session/{session}/execute/sync"
        payload = {"script": script, "args": args}
        response = __post(url, payload)
        return response.get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to run the script.") from error


def get_alert_text(driver_url, session):
    """Get the text from an alert"""
    try:
        url = f"{driver_url}/session/{session}/alert/text"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the alert text.") from error


def get_active_element(driver_url, session):
    """Get the active element"""
    try:
        url = f"{driver_url}/session/{session}/element/active"
        response = __get(url)
        return helper.get_element(response)
    except Exception as error:
        raise WebDriverError(f"Failed to get the active element.") from error


def clear_element(driver_url, session, element):
    """Clear the element text"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/clear"
        payload = {"id": element}
        __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to clear the element text.") from error


def is_element_enabled(driver_url, session, element):
    """Check if element is enabled"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/enabled"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to check if element is enabled.") from error


def get_css_value(driver_url, session, element, property_name):
    """Get the css property value"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/css/{property_name}"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get the css property value.") from error


def is_element_selected(driver_url, session, element):
    """Check if element is selected"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/selected"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to check if element is selected.") from error


def get_window_rectangle(driver_url, session):
    """Get window rectangle"""
    try:
        url = f"{driver_url}/session/{session}/window/rect"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get window rectangle.") from error


def get_window_handles(driver_url, session):
    """Get window handles"""
    try:
        url = f"{driver_url}/session/{session}/window/handles"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get window handles.") from error


def close_window(driver_url, session):
    """Close active window"""
    try:
        url = f"{driver_url}/session/{session}/window"
        return __delete(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to close active window.") from error


def get_window(driver_url, session):
    """Get window"""
    try:
        url = f"{driver_url}/session/{session}/window"
        return __get(url).get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get window.") from error


def go_back(driver_url, session):
    """
    This command causes the browser to traverse one step backward in the joint session history of the
    current browse. This is equivalent to pressing the back button in the browser.
    """
    try:
        url = f"{driver_url}/session/{session}/back"
        __post(url, {})
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to go back to page.") from error


def get_url(driver_url, session):
    """Return the URL from web page:"""
    try:
        url = f"{driver_url}/session/{session}/url"
        response = __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get page url.") from error


def get_timeouts(driver_url, session):
    """
    Return the configured timeouts:
        {"implicit": 0, "pageLoad": 300000, "script": 30000}
    """
    try:
        url = f"{driver_url}/session/{session}/timeouts"
        response = __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get timeouts.") from error


def get_status(driver_url):
    """
    Return the status and details of the WebDriver:
        "build": {
                "version": "113.0.5672.63 (0e1a4471d5ae5bf128b1bd8f4d627c8cbd55f70c-refs/branch-heads/5672@{#912})"
            },
            "message": "ChromeDriver ready for new sessions.",
            "os": {"arch": "x86_64", "name": "Linux", "version": "5.4.0-150-generic"},
            "ready": True,
        }
    """
    try:
        url = f"{driver_url}/status"
        return __get(url)
    except Exception as error:
        raise WebDriverError(f"Failed to get status.") from error


def get_title(driver_url, session):
    """Get the page title"""
    try:
        url = f"{driver_url}/session/{session}/title"
        response = __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError(f"Failed to get page title.") from error


def find_elements(driver_url, session, locator_type, locator_value):
    """Search the DOM elements by 'locator', for example, 'xpath'"""
    try:
        url = f"{driver_url}/session/{session}/elements"
        payload = {"using": locator_type, "value": locator_value}
        response = __post(url, payload)
        return [x.get("ELEMENT") for x in response.get("value")]
    except Exception as error:
        raise WebDriverError(
            f"Failed to find elements by '{locator_type}'-'{locator_value}'."
        ) from error


def get_property(driver_url, session, element, property):
    """Get the given HTML property of an element, for example, 'href'"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/property/{property}"
        response = __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get value from element.") from error


def get_attribute(driver_url, session, element, attribute):
    """Get the given HTML attribute of an element, for example, 'aria-valuenow'"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/attribute/{attribute}"
        response = __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get value from element.") from error


def get_cookies(driver_url, session):
    """Get the page cookies"""
    try:
        url = f"{driver_url}/session/{session}/cookie"
        response = __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get page cookies.") from error


def get(driver_url, session, page_url):
    """Does the same of 'go_to_page'. Added to be compatible with selenium method name'"""
    return go_to_page(driver_url, session, page_url)


def go_to_page(driver_url, session, page_url):
    """Navigate to 'page_url'"""
    try:
        url = f"{driver_url}/session/{session}/url"
        payload = {"url": page_url}
        __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to navigate to '{page_url}'") from error


def close_session(driver_url, session):
    """Close an opened session and close the browser"""
    try:
        url = f"{driver_url}/session/{session}"
        __delete(url)
        return True
    except Exception as error:
        raise WebDriverError("Failed to close session.") from error


def get_text(driver_url, session, element):
    """Get the text of an element"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/text"
        response = __get(url)
        return response.get("value")
    except Exception as error:
        raise WebDriverError("Failed to get text from element.") from error


def send_keys(driver_url, session, element, text):
    """Fill an editable element, for example a textarea, with a given text"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/value"
        payload = {"text": text, "value": [*text], "id": element}
        __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError(f"Failed to send key '{text}'.") from error


def click(driver_url, session, element):
    """Click on an element"""
    try:
        url = f"{driver_url}/session/{session}/element/{element}/click"
        payload = {"id": element}
        __post(url, payload)
        return True
    except Exception as error:
        raise WebDriverError("Failed to click on element.") from error


def __get_session(response):
    # Firefox response
    value = response.get("value")
    session_id = value.get("sessionId")
    if session_id:
        return session_id

    # Chrome response
    return response.get("sessionId")


def get_session(driver_url, capabilities):
    """Opens a browser and a session. This session is used for all functions to perform events in the page"""
    try:
        url = f"{driver_url}/session"
        data = capabilities
        response = __post(url, payload=data)
        return __get_session(response)
    except Exception as error:
        raise WebDriverError("Failed to open session.") from error


def find_element(driver_url, session, locator_type, locator_value):
    """Find an element by a 'locator', for example 'xpath'"""
    try:
        url = f"{driver_url}/session/{session}/element"
        payload = {"using": locator_type, "value": locator_value}
        response = __post(url, payload)

        # Firefox does not support id locator, so it prints the error message to the user
        # It helps on debug
        if response.get("value").get("error"):
            raise WebDriverError(f"Failed to find element. {response}")

        return helper.get_element(response)
    except Exception as error:
        raise WebDriverError(
            f"Failed to find element by '{locator_type}'-'{locator_value}'."
        ) from error



# All fake responses where collected from chromedriver responses


class Dictionary:
    def __init__(self, dictionary) -> None:
        self.dictionary = dictionary

    def json(self):
        return self.dictionary

    # used by sync functions
    def get(self, key, *args):
        return self.dictionary.get(key)


def dict_to_json(dictionary):
    class MockResponse:
        @property
        def status_code(self):
            return 200

        # used by async functions
        def get(self, argument, *args):
            return dictionary.get(argument)

        def json(self):
            return Dictionary(dictionary)

    return MockResponse()


DEFAULT = dict_to_json(
    {
        "sessionId": "4358a5b53794586af59678fc1653dc40",
        "status": 0,
        "value": {"ELEMENT": "0.8851292311864847-1"},
    }
)

FIND_ELEMENT = DEFAULT
SEND_KEYS = DEFAULT
CLICK = DEFAULT
CLOSE_SESSION = DEFAULT
GO_TO_PAGE = DEFAULT

GET_NAMED_COOKIE = dict_to_json(
    {
        "value": {
            "name": "username",
            "value": "John Doe",
            "path": "//home/user/fullpath",
            "domain": "",
            "secure": False,
            "httpOnly": False,
            "sameSite": "None",
        }
    }
)

GET_RECT = dict_to_json(
    {
        "sessionId": "a9d6e77726f3eda12e92b06b5066dbb4",
        "status": 0,
        "value": {"height": 23, "width": 183, "x": 10, "y": 9652.12},
    }
)

ACTIONS = dict_to_json(
    {"sessionId": "449dbd1df001e9a9e13b3bac5babe809", "status": 0, "value": "null"}
)

EXECUTE_SCRIPT = dict_to_json(
    {"sessionId": "9f4a4a9420663d0c0cc18957ab463b90", "status": 0, "value": "any"}
)

GET_PAGE_SOURCE = dict_to_json(
    {
        "sessionId": "e34234d1445ed6d4833370d1d8019282",
        "status": 0,
        "value": "<html><head><title>Sample page</title></head><body><h1>Basic page</h1></body></html>",
    }
)

GET_ALERT_TEXT = dict_to_json(
    {"sessionId": "171ba19c927e0b95e1a53dbbdcfcdc19", "status": 0, "value": "any warn"}
)

GET_WINDOW_RECTANGLE = dict_to_json(
    {
        "sessionId": "79e4bd950a886e0119e3760d201b059e",
        "status": 0,
        "value": {"height": 600, "width": 800, "x": 0, "y": 0},
    }
)

GET_ACTIVE_ELEMENT = DEFAULT

CLEAR_ELEMENT = dict_to_json(
    {"sessionId": "486fa32a9876b4519e149b39135edcb5", "status": 0, "value": None}
)

IS_ELEMENT_ENABLED = dict_to_json(
    {"sessionId": "e0e43cd1ce532b5aa62b6df0de11e3bd", "status": 0, "value": True}
)

GET_CSS_COLOR_VALUE = dict_to_json(
    {
        "sessionId": "e7f659e7183778e98ce9357051e40a47",
        "status": 0,
        "value": "rgba(0, 0, 0, 1)",
    }
)

IS_ELEMENT_SELECTED = dict_to_json(
    {"sessionId": "341d21063846141d2716d300652ddd81", "status": 0, "value": False}
)

GET_WINDOW_HANDLES = dict_to_json(
    {
        "sessionId": "b3f92cf70d734ecc6fcddbd88671998a",
        "status": 0,
        "value": ["2E55CCE389196328988ED244DAA52A5D"],
    }
)

CLOSE_WINDOW = dict_to_json(
    {"sessionId": "48399161591a0bc1dffaa2ff2d65aa0f", "status": 0, "value": []}
)

GET_WINDOW = dict_to_json(
    {
        "sessionId": "ce68162d420e9cb2b1617c2d1a800f85",
        "status": 0,
        "value": "845623CAE8115F2B60C9AE8596F13D94",
    }
)

GET_URL = dict_to_json(
    {
        "sessionId": "af67b8ef665d30a687f37365d229fb53",
        "status": 0,
        "value": "file:///html/playground.html",
    }
)
GET_TIMEOUTS = dict_to_json(
    {
        "sessionId": "10754c8ec2e19133235223f1914ea376",
        "status": 0,
        "value": {"implicit": 0, "pageLoad": 300000, "script": 30000},
    }
)


GET_STATUS = dict_to_json(
    {
        "value": {
            "build": {
                "version": "113.0.5672.63 (0e1a4471d5ae5bf128b1bd8f4d627c8cbd55f70c-refs/branch-heads/5672@{#912})"
            },
            "message": "ChromeDriver ready for new sessions.",
            "os": {"arch": "x86_64", "name": "Linux", "version": "5.4.0-150-generic"},
            "ready": True,
        },
        "status": "0",  # fake status fro mock purposes. Not present in real response
    }
)

GET_TITLE = dict_to_json(
    {
        "sessionId": "07b00b2e94be84920495d83890c82b60",
        "status": 0,
        "value": "Sample page",
    }
)

GET_COOKIES = dict_to_json(
    {
        "sessionId": "07b00b2e94be84920495d83890c82b60",
        "status": 0,
        "value": [],
    }
)


FIND_ELEMENTS = dict_to_json(
    {
        "sessionId": "9be93a374d185216134bf0c3fafee52e",
        "status": 0,
        "value": [
            {"ELEMENT": "C230605181E69CB2C4C36B8E83FE1245_element_1"},
            {"ELEMENT": "C230605181E69CB2C4C36B8E83FE1245_element_2"},
            {"ELEMENT": "C230605181E69CB2C4C36B8E83FE1245_element_3"},
        ],
    }
)

GET_PROPERTY_VALUE = dict_to_json(
    {
        "sessionId": "5be82d4cd17af92d7ea53a36900d78cb",
        "status": 0,
        "value": "any_value",
    }
)

GET_ATTRIBUTE_VALUE = dict_to_json(
    {
        "sessionId": "5be82d4cd17af92d7ea53a36900d78cb",
        "status": 0,
        "value": "any_value",
    }
)

GET_TEXT = dict_to_json(
    {"sessionId": "5be82d4cd17af92d7ea53a36900d78cb", "status": 0, "value": "any"}
)

GET_SESSION = dict_to_json(
    {
        "sessionId": "4358a5b53794586af59678fc1653dc40",
        "status": 0,
        "value": {
            "acceptInsecureCerts": True,
            "acceptSslCerts": True,
            "applicationCacheEnabled": False,
            "browserConnectionEnabled": False,
            "browserName": "chrome",
            "chrome": {
                "chromedriverVersion": "94.0.4606.41 (333e85df3c9b656b518b5f1add5ff246365b6c24-refs/branch-heads/4606@{#845})",
                "userDataDir": "/tmp/.com.google.Chrome.4zKpeQ",
            },
            "cssSelectorsEnabled": True,
            "databaseEnabled": False,
            "goog:chromeOptions": {"debuggerAddress": "localhost:43565"},
            "handlesAlerts": True,
            "hasTouchScreen": False,
            "javascriptEnabled": True,
            "locationContextEnabled": True,
            "mobileEmulationEnabled": False,
            "nativeEvents": True,
            "networkConnectionEnabled": False,
            "pageLoadStrategy": "normal",
            "platform": "Linux",
            "proxy": {},
            "rotatable": False,
            "setWindowRect": True,
            "strictFileInteractability": False,
            "takesHeapSnapshot": True,
            "takesScreenshot": True,
            "timeouts": {"implicit": 0, "pageLoad": 300000, "script": 30000},
            "unexpectedAlertBehaviour": "ignore",
            "version": "94.0.4606.54",
            "webStorageEnabled": True,
            "webauthn:extension:credBlob": True,
            "webauthn:extension:largeBlob": True,
            "webauthn:virtualAuthenticators": True,
        },
    }
)



# It opens the WebDriver, navigate to a page and get all links
import asyncio
import time
from caqui import synchronous, asynchronous
from os import getcwd
from tests.constants import PAGE_URL

BASE_DIR = getcwd()

MAX_CONCURRENCY = 5  # number of webdriver instances running
all_anchors = []
semaphore = asyncio.Semaphore(MAX_CONCURRENCY)


async def get_all_links():
    async with semaphore:
        driver_url = "http://127.0.0.1:9999"
        capabilities = {
            "desiredCapabilities": {
                "browserName": "chrome",
                "marionette": True,
                "acceptInsecureCerts": True,
                "pageLoadStrategy": "normal",
                "goog:chromeOptions": {"extensions": [], "args": ["--headless"]},
            }
        }

        session = await asynchronous.get_session(driver_url, capabilities)
        await asynchronous.go_to_page(
            driver_url,
            session,
            PAGE_URL,
        )

        all_anchors = []
        for i in range(4):
            i += 1
            locator_value = f"//a[@id='a{i}']"
            locator_type = "xpath"
            anchors = []

            anchors = await asynchronous.find_elements(
                driver_url, session, locator_type, locator_value
            )
            all_anchors.extend(anchors)

        texts = []
        for anchor in all_anchors:
            text = await asynchronous.get_property(driver_url, session, anchor, "href")
            texts.append(text)

        for text in texts:
            print(f"Link found '{text}'")

        synchronous.close_session(driver_url, session)


# Reference: https://stackoverflow.com/questions/48483348/how-to-limit-concurrency-with-python-asyncio
async def main():
    number_of_websites = range(10)
    tasks = [asyncio.ensure_future(get_all_links()) for number in number_of_websites]
    await asyncio.gather(*tasks)


if __name__ == "__main__":
    start = time.time()
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(main())
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()
        end = time.time()
        print(f"Found 40 links")  # 10 websites with 4 links each
        print(f"Time: {end-start:.2f} sec")



from caqui.helper import get_elements, get_element
from pytest import mark
from tests.fake_responses import FIND_ELEMENTS, FIND_ELEMENT


@mark.parametrize(
    "response,expected",
    [
        (
            {
                "value": {
                    "element-6066-11e4-a52e-4f735466cecf": "c4cab128-a0a4-4355-93a5-ffd6c7a8b042"
                },
            },
            "c4cab128-a0a4-4355-93a5-ffd6c7a8b042",
        ),
        (
            FIND_ELEMENT,
            "0.8851292311864847-1",
        ),
    ],
)
def test_get_element(response, expected):
    assert get_element(response) == expected


@mark.parametrize(
    "response,expected",
    [
        (
            {
                "value": [
                    {
                        "element-6066-11e4-a52e-4f735466cecf": "c4cab128-a0a4-4355-93a5-ffd6c7a8b042"
                    },
                    {
                        "element-6066-11e4-a52e-4f735466cecf": "12345678-a0a4-4355-93a5-ffd6c7a8b042"
                    },
                ]
            },
            [
                "c4cab128-a0a4-4355-93a5-ffd6c7a8b042",
                "12345678-a0a4-4355-93a5-ffd6c7a8b042",
            ],
        ),
        (
            FIND_ELEMENTS,
            [
                "C230605181E69CB2C4C36B8E83FE1245_element_1",
                "C230605181E69CB2C4C36B8E83FE1245_element_2",
                "C230605181E69CB2C4C36B8E83FE1245_element_3",
            ],
        ),
    ],
)
def test_get_elements(response, expected):
    assert get_elements(response) == expected



# Simple example of usage of caqui with Windows Calculator
# It opens the Calculator and clicks the number "8"
# Test works just in Windows environment. Tested with Windows 10
from caqui import synchronous


def main():
    driver_url = "http://127.0.0.1:9999"
    capabilities = {
        "capabilities": {"firstMatch": [{}], "alwaysMatch": {}},
        "desiredCapabilities": {
            "debugConnectToRunningApp": "false",
            "app": "C:/windows/system32/calc.exe",
        },
    }
    session = synchronous.get_session(driver_url, capabilities)
    element = synchronous.find_element(
        driver_url, session, locator_type="name", locator_value="Eight"
    )

    synchronous.click(driver_url, session, element)
    synchronous.close_session(driver_url, session)


if __name__ == "__main__":
    main()



# necessary import for mokcs
import caqui

from pytest import mark
from caqui import asynchronous
from tests import fake_responses
from unittest.mock import patch


async def mock_request(*args):
    pass


@mark.asyncio
async def test_get_rect():
    expected = {"height": 23, "width": 183, "x": 10, "y": 9652.12}

    async def mock_request(*args):
        return fake_responses.GET_RECT

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_rect("", "", "") == expected


@mark.asyncio
async def test_actions_scroll_to_element():
    async def mock_request(*args):
        return fake_responses.ACTIONS

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.actions_scroll_to_element("", "", "") == True


@mark.asyncio
async def test_submit():
    async def mock_request(*args):
        return fake_responses.CLICK

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.submit("", "", "") == True


@mark.asyncio
async def test_actions_click():
    async def mock_request(*args):
        return fake_responses.ACTIONS

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.actions_click("", "", "") == True


@mark.asyncio
async def test_set_timeouts():
    async def mock_request(*args):
        return fake_responses.GET_TIMEOUTS

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.set_timeouts("", "", "") == True


@mark.asyncio
async def test_find_children_elements():
    element = "C230605181E69CB2C4C36B8E83FE1245_element_2"

    async def mock_request(*args):
        return fake_responses.FIND_ELEMENTS

    with patch("caqui.asynchronous.__post", mock_request):
        assert element in await asynchronous.find_children_elements("", "", "", "", "")


@mark.asyncio
async def test_find_child_element():
    element = "0.8851292311864847-1"

    async def mock_request(*args):
        return fake_responses.FIND_ELEMENT

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.find_child_element("", "", "", "", "") == element


@mark.asyncio
async def test_execute_script():
    expected = "any"

    async def mock_request(*args):
        return fake_responses.EXECUTE_SCRIPT

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.execute_script("", "", "", "") == expected


@mark.asyncio
async def test_get_page_source():
    expected = "Sample page"

    async def mock_request(*args):
        return fake_responses.GET_PAGE_SOURCE

    with patch("caqui.asynchronous.__get", mock_request):
        assert expected in await asynchronous.get_page_source("", "")


@mark.asyncio
async def test_get_alert_text():
    expected = "any warn"

    async def mock_request(*args):
        return fake_responses.GET_ALERT_TEXT

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_alert_text("", "") == expected


@mark.asyncio
async def test_get_active_element():
    expected = "0.8851292311864847-1"

    async def mock_request(*args):
        return fake_responses.GET_ACTIVE_ELEMENT

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_active_element("", "") == expected


@mark.asyncio
async def test_clear_element():
    async def mock_request(*args):
        return fake_responses.CLEAR_ELEMENT

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.clear_element("", "", "") is True


@mark.asyncio
async def test_is_element_enabled():
    async def mock_request(*args):
        return fake_responses.IS_ELEMENT_ENABLED

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.is_element_enabled("", "", "") is True


@mark.asyncio
async def test_get_css_value():
    expected = "rgba(0, 0, 0, 1)"

    async def mock_request(*args):
        return fake_responses.GET_CSS_COLOR_VALUE

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_css_value("", "", "", "") == expected


@mark.asyncio
async def test_is_element_selected():
    async def mock_request(*args):
        return fake_responses.IS_ELEMENT_SELECTED

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.is_element_selected("", "", "") is False


@mark.asyncio
async def test_get_window_rectangle():
    expected = "height"

    async def mock_request(*args):
        return fake_responses.GET_WINDOW_RECTANGLE

    with patch("caqui.asynchronous.__get", mock_request):
        assert expected in await asynchronous.get_window_rectangle("", "")


@mark.asyncio
async def test_get_window_handles():
    expected = ["2E55CCE389196328988ED244DAA52A5D"]

    async def mock_request(*args):
        return fake_responses.GET_WINDOW_HANDLES

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_window_handles("", "") == expected


@mark.asyncio
async def test_close_window():
    expected = []

    async def mock_request(*args):
        return fake_responses.CLOSE_WINDOW

    with patch("caqui.asynchronous.__delete", mock_request):
        assert await asynchronous.close_window("", "") == expected


@mark.asyncio
async def test_get_window():
    expected = "845623CAE8115F2B60C9AE8596F13D94"

    async def mock_request(*args):
        return fake_responses.GET_WINDOW

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_window("", "") == expected


@mark.asyncio
async def test_go_back():
    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.go_back("", "") is True


@mark.asyncio
async def test_get_property():
    expected = "any_value"

    async def mock_request(*args):
        return fake_responses.GET_PROPERTY_VALUE

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_property("", "", "", "") == expected


@mark.asyncio
async def test_get_attribute():
    expected = "any_value"

    async def mock_request(*args):
        return fake_responses.GET_ATTRIBUTE_VALUE

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_attribute("", "", "", "") == expected


@mark.asyncio
async def test_get_url():
    expected = "playground.html"

    async def mock_request(*args):
        return fake_responses.GET_URL

    with patch("caqui.asynchronous.__get", mock_request):
        response = await asynchronous.get_url("", "")
        assert expected in response


@mark.asyncio
async def test_get_timeouts():
    expected = "implicit"

    async def mock_request(*args):
        return fake_responses.GET_TIMEOUTS

    with patch("caqui.asynchronous.__get", mock_request):
        response = await asynchronous.get_timeouts("", "")
        assert expected in response


@mark.asyncio
async def test_get_status():
    async def mock_request(*args):
        return fake_responses.GET_STATUS

    with patch("caqui.asynchronous.__get", mock_request):
        response = await asynchronous.get_status("")
        assert response.get("value").get("ready") is True


@mark.asyncio
async def test_get_title():
    expected = "Sample page"

    async def mock_request(*args):
        return fake_responses.GET_TITLE

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_title("", "") == expected


@mark.asyncio
async def test_get_cookies():
    expected = []

    async def mock_request(*args):
        return fake_responses.GET_COOKIES

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_cookies("", "") == expected


@mark.asyncio
async def test_get_text():
    expected = "any"

    async def mock_request(*args):
        return fake_responses.GET_TEXT

    with patch("caqui.asynchronous.__get", mock_request):
        assert await asynchronous.get_text("", "", "") == expected


@mark.asyncio
async def test_close_session():
    with patch("caqui.asynchronous.__delete", mock_request):
        assert await asynchronous.close_session("", "") is True


@mark.asyncio
async def test_go_to_page():
    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.go_to_page("", "", "") is True


@mark.asyncio
async def test_send_keys():
    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.send_keys("", "", "", "") is True


@mark.asyncio
async def test_click():
    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.click("", "", "") is True


@mark.asyncio
async def test_find_elements():
    element = "C230605181E69CB2C4C36B8E83FE1245_element_2"

    async def mock_request(*args):
        return fake_responses.FIND_ELEMENTS

    with patch("caqui.asynchronous.__post", mock_request):
        assert element in await asynchronous.find_elements("", "", "", "")


@mark.asyncio
async def test_find_element():
    element = "0.8851292311864847-1"

    async def mock_request(*args):
        return fake_responses.FIND_ELEMENT

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.find_element("", "", "", "") == element


@mark.asyncio
async def test_get_session():
    expected = "4358a5b53794586af59678fc1653dc40"

    async def mock_request(*args):
        return fake_responses.GET_SESSION

    with patch("caqui.asynchronous.__post", mock_request):
        assert await asynchronous.get_session("", {}) == expected



from caqui.synchronous import (
    find_element,
    get_session,
    click,
    send_keys,
    get_text,
    close_session,
    go_to_page,
    get_property,
    clear_element,
    get_rect,
    get_css_value,
    get_attribute,
    switch_to_frame,
    switch_to_parent_frame,
    dismiss_alert,
)
from tests.constants import PAGE_URL
from pytest import fixture


@fixture
def __setup():
    driver_url = "http://127.0.0.1:9999"
    capabilities = {
        "desiredCapabilities": {
            "browserName": "chrome",
            "marionette": True,
            "acceptInsecureCerts": True,
            "goog:chromeOptions": {"extensions": [], "args": ["--headless"]},
        }
    }
    session = get_session(driver_url, capabilities)
    go_to_page(
        driver_url,
        session,
        PAGE_URL,
    )
    yield driver_url, session
    close_session(driver_url, session)


def test_switch_to_parent_frame_and_click_alert(__setup):
    driver_url, session = __setup
    locator_type = "id"
    locator_value = "my-iframe"
    locator_value_alert_parent = "alert-button"
    locator_value_alert_frame = "alert-button-iframe"

    element_frame = find_element(driver_url, session, locator_type, locator_value)
    assert switch_to_frame(driver_url, session, element_frame) is True

    alert_button_frame = find_element(
        driver_url, session, locator_type, locator_value_alert_frame
    )
    assert click(driver_url, session, alert_button_frame) is True
    assert dismiss_alert(driver_url, session) is True

    assert switch_to_parent_frame(driver_url, session, element_frame) is True
    alert_button_parent = find_element(
        driver_url, session, locator_type, locator_value_alert_parent
    )
    assert get_attribute(driver_url, session, alert_button_parent, "any") == "any"
    assert click(driver_url, session, alert_button_parent) is True


def test_switch_to_frame_and_click_alert(__setup):
    driver_url, session = __setup
    locator_type = "id"
    locator_value = "my-iframe"
    locator_value_alert = "alert-button-iframe"

    element_frame = find_element(driver_url, session, locator_type, locator_value)
    assert switch_to_frame(driver_url, session, element_frame) is True

    alert_button = find_element(driver_url, session, locator_type, locator_value_alert)
    assert get_attribute(driver_url, session, alert_button, "any") == "any"
    assert click(driver_url, session, alert_button) is True


def test_get_data_from_hidden_button(__setup):
    driver_url, session = __setup
    locator_type = "xpath"

    hidden_button = find_element(
        driver_url, session, locator_type, locator_value="//*[@id='hidden-button']"
    )

    assert "width" in get_rect(driver_url, session, hidden_button)
    assert "visible" == get_css_value(driver_url, session, hidden_button, "visibility")
    assert True == get_property(driver_url, session, hidden_button, "hidden")
    assert ["display"] == get_property(driver_url, session, hidden_button, "style")
    assert "display: none;" == get_attribute(
        driver_url, session, hidden_button, "style"
    )


def test_add_text__click_button_and_get_properties(__setup):
    driver_url, session = __setup
    expected = "end"
    locator_type = "xpath"

    input = find_element(driver_url, session, locator_type, locator_value="//input")
    send_keys(driver_url, session, input, "any")
    assert get_property(driver_url, session, input, property="value") == "any"
    clear_element(driver_url, session, input)
    assert get_property(driver_url, session, input, property="value") == ""

    anchor = find_element(driver_url, session, locator_type, locator_value="//a")
    assert (
        get_property(driver_url, session, anchor, property="href") == "http://any1.com/"
    )

    button = find_element(driver_url, session, locator_type, locator_value="//button")
    click(driver_url, session, button)

    p = find_element(driver_url, session, locator_type, locator_value="//p[@id='end']")
    get_text(driver_url, session, p)

    assert get_text(driver_url, session, p) == expected



import base64


def save_picture(session, path, file_name, response):
    with open(f"{path}/{file_name}-{session}.png", "wb") as f:
        f.write(base64.b64decode((response)))


def get_elements(response):
    values = response.get("value")
    return [list(value.values())[0] for value in values]


def get_element(response):
    value = response.get("value")
    # Google Chrome
    element = value.get("ELEMENT")
    if element:
        return element

    # Firefox
    return list(value.values())[0]



from pytest import fixture, mark, raises
from caqui import asynchronous, synchronous
from tests.constants import PAGE_URL
from caqui.exceptions import WebDriverError


@fixture
def __setup():
    driver_url = "http://127.0.0.1:9999"
    capabilities = {
        "desiredCapabilities": {
            "name": "webdriver",
            "browserName": "chrome",
            "marionette": False,
            "acceptInsecureCerts": True,
            "goog:chromeOptions": {"extensions": [], "args": ["--headless"]},
        }
    }
    session = synchronous.get_session(driver_url, capabilities)
    synchronous.go_to_page(
        driver_url,
        session,
        PAGE_URL,
    )
    yield driver_url, session
    synchronous.close_session(driver_url, session)


@mark.asyncio
async def test_switch_to_parent_frame_asynchronous(__setup):
    driver_url, session = __setup
    locator_type = "id"
    locator_value = "my-iframe"

    element_frame = synchronous.find_element(
        driver_url, session, locator_type, locator_value
    )
    assert (
        await asynchronous.switch_to_parent_frame(driver_url, session, element_frame)
        is True
    )


def test_switch_to_parent_frame_synchronous(__setup):
    driver_url, session = __setup
    locator_type = "id"
    locator_value = "my-iframe"

    element_frame = synchronous.find_element(
        driver_url, session, locator_type, locator_value
    )
    assert (
        synchronous.switch_to_parent_frame(driver_url, session, element_frame) is True
    )


@mark.asyncio
async def test_switch_to_frame_asynchronous(__setup):
    driver_url, session = __setup
    locator_type = "id"
    locator_value = "my-iframe"

    element_frame = synchronous.find_element(
        driver_url, session, locator_type, locator_value
    )
    assert (
        await asynchronous.switch_to_frame(driver_url, session, element_frame) is True
    )


def test_switch_to_frame_synchronous(__setup):
    driver_url, session = __setup
    locator_type = "id"
    locator_value = "my-iframe"

    element_frame = synchronous.find_element(
        driver_url, session, locator_type, locator_value
    )
    assert synchronous.switch_to_frame(driver_url, session, element_frame) is True


@mark.asyncio
async def test_send_alert_text(__setup):
    driver_url, session = __setup
    locator_type = "css selector"
    locator_value = "#alert-button-prompt"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    synchronous.click(driver_url, session, element)

    assert synchronous.send_alert_text(driver_url, session, text="any1") is True
    synchronous.accept_alert(driver_url, session) is True

    synchronous.click(driver_url, session, element)
    assert await asynchronous.send_alert_text(driver_url, session, "any2") is True
    synchronous.accept_alert(driver_url, session) is True


@mark.asyncio
async def test_accept_alert(__setup):
    driver_url, session = __setup
    locator_type = "css selector"
    locator_value = "#alert-button"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    synchronous.click(driver_url, session, element)

    assert synchronous.accept_alert(driver_url, session) is True

    synchronous.click(driver_url, session, element)
    assert await asynchronous.accept_alert(driver_url, session) is True


@mark.asyncio
async def test_dismiss_alert(__setup):
    driver_url, session = __setup
    locator_type = "css selector"
    locator_value = "#alert-button"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    synchronous.click(driver_url, session, element)

    assert synchronous.dismiss_alert(driver_url, session) is True

    synchronous.click(driver_url, session, element)
    assert await asynchronous.dismiss_alert(driver_url, session) is True


@mark.asyncio
async def test_take_screenshot_element(__setup):
    driver_url, session = __setup
    locator_type = "css selector"
    locator_value = "#alert-button"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert synchronous.take_screenshot_element(driver_url, session, element) is True
    assert (
        await asynchronous.take_screenshot_element(driver_url, session, element) is True
    )


@mark.asyncio
async def test_take_screenshot(__setup):
    driver_url, session = __setup

    assert synchronous.take_screenshot(driver_url, session) is True
    assert await asynchronous.take_screenshot(driver_url, session) is True


@mark.asyncio
async def test_delete_cookies_asynchronous(__setup):
    driver_url, session = __setup

    response = await asynchronous.delete_all_cookies(driver_url, session)
    assert response is True


@mark.asyncio
async def test_delete_cookies_synchronous(__setup):
    driver_url, session = __setup

    assert synchronous.delete_all_cookies(driver_url, session) is True


@mark.skip(reason="works just with Firefox")
@mark.asyncio
async def test_get_named_cookie(__setup):
    driver_url, session = __setup
    name = "username"  # cookie created on page load
    expected = "John Doe"

    assert (
        synchronous.get_named_cookie(driver_url, session, name).get("value") == expected
    )
    response = await asynchronous.get_named_cookie(driver_url, session, name)
    assert response.get("value") == expected


@mark.asyncio
async def test_get_computed_label(__setup):
    driver_url, session = __setup
    locator_type = "css selector"
    locator_value = "#alert-button"
    expected = "alert"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert synchronous.get_computed_label(driver_url, session, element) == expected

    assert (
        await asynchronous.get_computed_label(driver_url, session, element) == expected
    )


@mark.asyncio
async def test_get_computed_role(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"
    expected = "textbox"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert synchronous.get_computed_role(driver_url, session, element) == expected

    assert (
        await asynchronous.get_computed_role(driver_url, session, element) == expected
    )


@mark.asyncio
async def test_get_tag_name(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"
    expected = "input"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert synchronous.get_tag_name(driver_url, session, element) == expected

    assert await asynchronous.get_tag_name(driver_url, session, element) == expected


@mark.asyncio
async def test_get_shadow_root(__setup):
    driver_url, session = __setup
    locator_type = "id"
    locator_value = "shadow-root"
    root_element = "shadow-6066-11e4-a52e-4f735466cecf"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert (
        synchronous.get_shadow_root(driver_url, session, element).get(root_element)
        is not None
    )

    response = await asynchronous.get_shadow_root(driver_url, session, element)
    assert response.get(root_element) is not None


@mark.asyncio
async def test_get_rect(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"
    expected = {"height": 21, "width": 185, "x": 8, "y": 100.4375}

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert synchronous.get_rect(driver_url, session, element) == expected

    assert await asynchronous.get_rect(driver_url, session, element) == expected


@mark.asyncio
async def test_actions_scroll_to_element(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//button"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    assert synchronous.actions_scroll_to_element(driver_url, session, element) is True
    assert (
        await asynchronous.actions_scroll_to_element(driver_url, session, element)
        is True
    )


@mark.asyncio
async def test_submit(__setup):
    driver_url, session = __setup
    locator_type = "name"
    locator_value = "my-form"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    assert synchronous.submit(driver_url, session, element) is True

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    assert await asynchronous.submit(driver_url, session, element) is True


@mark.asyncio
async def test_actions_click(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//button"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    assert synchronous.actions_click(driver_url, session, element) is True
    assert await asynchronous.actions_click(driver_url, session, element) is True


@mark.asyncio
async def test_raise_exception_when_element_not_found(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//invalid-tag"

    with raises(WebDriverError):
        synchronous.find_element(driver_url, session, locator_type, locator_value)

    with raises(WebDriverError):
        await asynchronous.find_element(
            driver_url, session, locator_type, locator_value
        )


@mark.asyncio
async def test_set_timeouts(__setup):
    driver_url, session = __setup
    timeouts_1 = 5000  # milliseconds
    timeouts_2 = 3000  # milliseconds

    synchronous.set_timeouts(driver_url, session, timeouts_1)

    assert synchronous.get_timeouts(driver_url, session).get("implicit") == timeouts_1

    await asynchronous.set_timeouts(driver_url, session, timeouts_2)

    assert synchronous.get_timeouts(driver_url, session).get("implicit") == timeouts_2


@mark.asyncio
async def test_find_children_elements(__setup):
    driver_url, session = __setup
    expected = 1  # parent inclusive
    locator_type = "xpath"
    locator_value = "//div"

    parent_element = synchronous.find_element(
        driver_url, session, locator_type, '//div[@class="parent"]'
    )

    children_elements = synchronous.find_children_elements(
        driver_url, session, parent_element, locator_type, locator_value
    )

    assert len(children_elements) > expected

    children_elements = await asynchronous.find_children_elements(
        driver_url, session, parent_element, locator_type, locator_value
    )

    assert len(children_elements) > expected


@mark.asyncio
async def test_find_child_element(__setup):
    driver_url, session = __setup
    expected = "any4"
    locator_type = "xpath"
    locator_value = '//div[@class="child4"]'

    parent_element = synchronous.find_element(
        driver_url, session, locator_type, '//div[@class="parent"]'
    )

    child_element = synchronous.find_child_element(
        driver_url, session, parent_element, locator_type, locator_value
    )

    text = synchronous.get_text(driver_url, session, child_element)

    assert text == expected
    child_element = await asynchronous.find_child_element(
        driver_url, session, parent_element, locator_type, locator_value
    )
    text = synchronous.get_text(driver_url, session, child_element)
    assert text == expected


@mark.asyncio
async def test_get_page_source(__setup):
    driver_url, session = __setup
    expected = "Sample page"

    assert expected in synchronous.get_page_source(driver_url, session)
    assert expected in await asynchronous.get_page_source(driver_url, session)


@mark.asyncio
async def test_execute_script_asynchronous(__setup):
    driver_url, session = __setup
    script = "alert('any warn')"

    assert await asynchronous.execute_script(driver_url, session, script) == None


def test_execute_script_synchronous(__setup):
    driver_url, session = __setup
    script = "alert('any warn')"

    assert synchronous.execute_script(driver_url, session, script) == None


@mark.asyncio
async def test_get_alert_text(__setup):
    driver_url, session = __setup
    locator_type = "css selector"
    locator_value = "#alert-button"
    expected = "any warn"

    alert_button = synchronous.find_element(
        driver_url, session, locator_type, locator_value
    )
    synchronous.click(driver_url, session, alert_button)

    assert synchronous.get_alert_text(driver_url, session) == expected
    assert await asynchronous.get_alert_text(driver_url, session) == expected


@mark.asyncio
async def test_get_active_element(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    synchronous.send_keys(driver_url, session, element, "any")

    assert synchronous.get_active_element(driver_url, session) == element
    assert await asynchronous.get_active_element(driver_url, session) == element


@mark.asyncio
async def test_clear_element_fails_when_invalid_inputs(__setup):
    driver_url, session = __setup
    text = "any"
    element = "invalid"

    with raises(WebDriverError):
        synchronous.clear_element(driver_url, session, element) is True

    with raises(WebDriverError):
        await asynchronous.clear_element(driver_url, session, element)


@mark.asyncio
async def test_clear_element(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"
    text = "any"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    synchronous.send_keys(driver_url, session, element, text)
    assert synchronous.clear_element(driver_url, session, element) is True

    synchronous.send_keys(driver_url, session, element, text)
    assert await asynchronous.clear_element(driver_url, session, element) is True


@mark.asyncio
async def test_is_element_enabled(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert synchronous.is_element_enabled(driver_url, session, element) is True
    assert await asynchronous.is_element_enabled(driver_url, session, element) is True


@mark.asyncio
async def test_get_css_value(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"
    property_name = "color"
    expected = "rgba(0, 0, 0, 1)"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert (
        synchronous.get_css_value(driver_url, session, element, property_name)
        == expected
    )
    assert (
        await asynchronous.get_css_value(driver_url, session, element, property_name)
        == expected
    )


@mark.asyncio
async def test_is_element_selected(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert synchronous.is_element_selected(driver_url, session, element) is False
    assert await asynchronous.is_element_selected(driver_url, session, element) is False


@mark.asyncio
async def test_get_window_rectangle(__setup):
    driver_url, session = __setup
    expected = "height"

    assert expected in synchronous.get_window_rectangle(driver_url, session)
    rectangle = await asynchronous.get_window_rectangle(driver_url, session)
    assert expected in rectangle


@mark.asyncio
async def test_get_window_handles(__setup):
    driver_url, session = __setup

    assert isinstance(synchronous.get_window_handles(driver_url, session), list)
    handles = await asynchronous.get_window_handles(driver_url, session)
    assert isinstance(handles, list)


def test_close_window_sync(__setup):
    driver_url, session = __setup
    assert isinstance(synchronous.close_window(driver_url, session), list)


@mark.asyncio
async def test_close_window_async(__setup):
    driver_url, session = __setup

    response = await asynchronous.close_window(driver_url, session)
    assert isinstance(response, list)


@mark.asyncio
async def test_get_window(__setup):
    driver_url, session = __setup

    assert synchronous.get_window(driver_url, session) is not None
    assert await asynchronous.get_window(driver_url, session) is not None


@mark.asyncio
async def test_get_attribute_fails_when_invalid_attribute(__setup):
    driver_url, session = __setup
    attribute = "href"
    element = "invalid"

    with raises(WebDriverError):
        synchronous.get_attribute(driver_url, session, element, attribute)

    with raises(WebDriverError):
        await asynchronous.get_attribute(driver_url, session, element, attribute)


@mark.asyncio
async def test_get_attribute(__setup):
    driver_url, session = __setup
    attribute = "href"
    element = synchronous.find_element(driver_url, session, "xpath", "//a[@id='a1']")

    assert (
        synchronous.get_attribute(driver_url, session, element, attribute)
        == "http://any1.com/"
    )
    assert (
        await asynchronous.get_attribute(driver_url, session, element, attribute)
        == "http://any1.com/"
    )


@mark.asyncio
async def test_get_cookies(__setup):
    driver_url, session = __setup
    assert isinstance(synchronous.get_cookies(driver_url, session), list)
    cookies = await asynchronous.get_cookies(driver_url, session)
    assert isinstance(cookies, list)


@mark.asyncio
async def test_go_back(__setup):
    driver_url, session = __setup

    assert synchronous.go_back(driver_url, session) is True
    assert await asynchronous.go_back(driver_url, session) is True


@mark.asyncio
async def test_get_url(__setup):
    driver_url, session = __setup
    expected = "playground.html"

    assert expected in synchronous.get_url(driver_url, session)
    assert expected in await asynchronous.get_url(driver_url, session)


@mark.asyncio
async def test_get_timeouts(__setup):
    driver_url, session = __setup
    expected = "implicit"

    assert expected in synchronous.get_timeouts(driver_url, session)
    assert expected in await asynchronous.get_timeouts(driver_url, session)


@mark.asyncio
async def test_get_status(__setup):
    driver_url, _ = __setup
    expected = "ready"
    assert expected in synchronous.get_status(driver_url).get("value")
    response = await asynchronous.get_status(driver_url)
    assert expected in response.get("value")


@mark.asyncio
async def test_get_title(__setup):
    driver_url, session = __setup
    expected = "Sample page"

    assert synchronous.get_title(driver_url, session) == expected
    assert await asynchronous.get_title(driver_url, session) == expected


@mark.asyncio
async def test_find_elements_fails_when_invalid_data_input(__setup):
    driver_url, session = __setup
    locator_type = "invalid"
    locator_value = "//input"

    with raises(WebDriverError):
        synchronous.find_elements(driver_url, session, locator_type, locator_value)

    with raises(WebDriverError):
        await asynchronous.find_elements(
            driver_url, session, locator_type, locator_value
        )


@mark.asyncio
async def test_find_elements(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"

    elements = synchronous.find_elements(
        driver_url, session, locator_type, locator_value
    )
    async_elements = await asynchronous.find_elements(
        driver_url, session, locator_type, locator_value
    )

    assert len(elements) > 0
    assert len(async_elements) > 0


@mark.asyncio
async def test_find_element_fails_when_invalid_data_input(__setup):
    driver_url, session = __setup
    locator_type = "invalid"
    locator_value = "//input"

    with raises(WebDriverError):
        synchronous.find_element(driver_url, session, locator_type, locator_value)

    with raises(WebDriverError):
        await asynchronous.find_element(
            driver_url, session, locator_type, locator_value
        )


@mark.asyncio
async def test_find_element(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//input"

    assert (
        synchronous.find_element(driver_url, session, locator_type, locator_value)
        is not None
    )
    assert (
        await asynchronous.find_element(
            driver_url, session, locator_type, locator_value
        )
        is not None
    )


@mark.asyncio
async def test_get_property(__setup):
    driver_url, session = __setup
    text = "any_value"
    locator_type = "xpath"
    locator_value = "//input"
    property = "value"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)
    synchronous.send_keys(driver_url, session, element, text)

    assert synchronous.get_property(driver_url, session, element, property) == text
    assert (
        await asynchronous.get_property(driver_url, session, element, property) == text
    )


@mark.asyncio
async def test_get_text(__setup):
    driver_url, session = __setup
    expected = "end"
    locator_type = "xpath"
    locator_value = "//p[@id='end']"  # <p>end</p>

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert await asynchronous.get_text(driver_url, session, element) == expected
    assert synchronous.get_text(driver_url, session, element) == expected


@mark.asyncio
async def test_send_keys(__setup):
    driver_url, session = __setup
    text_async = "any_async"
    text_sync = "any_sync"
    locator_type = "xpath"
    locator_value = "//input"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert (
        await asynchronous.send_keys(driver_url, session, element, text_async) is True
    )
    assert synchronous.send_keys(driver_url, session, element, text_sync) is True


@mark.asyncio
async def test_click(__setup):
    driver_url, session = __setup
    locator_type = "xpath"
    locator_value = "//button"

    element = synchronous.find_element(driver_url, session, locator_type, locator_value)

    assert await asynchronous.click(driver_url, session, element) is True
    assert synchronous.click(driver_url, session, element) is True
